#include "path.h"
static LandQueue * _q;
static int _cmp_cb(void * data1, void * data2);
static int _estimate_distance(PathData * pd, int x1, int y1);
static int _estimate_unknown(PathData * pd, int x, int y);
static void _move(int * x, int * y, int d);
static int _cmp_cb(void * data1, void * data2) {
    PathNode * n1 = data1;
    PathNode * n2 = data2;
    if (n1->steps + n1->estimate < n2->steps + n2->estimate) {
        return - 1;
    }
    if (n1->steps + n1->estimate > n2->steps + n2->estimate) {
        return 1;
    }
    return 0;
}
static int _estimate_distance(PathData * pd, int x1, int y1) {
    return pow(pow(pd->x2 - x1, 2) + pow(pd->y2 - y1, 2), 0.5);
}
static int _estimate_unknown(PathData * pd, int x, int y) {
    return 0;
}
static void _move(int * x, int * y, int d) {
    if (d == 0) {
        * x += 1;
    }
    else if (d == 1) {
        * y -= 1;
    }
    else if (d == 2) {
        * x -= 1;
    }
    else if (d == 3) {
        * y += 1;
    }
}
int path_walk_road_cb(PathData * pd, int x, int y) {
    Tile * t = map_get(x, y);
    if (t->oid == 0) {
        if (pd->x2 == x && pd->y2 == y) {
            return 2;
        }
    }
    else if (t->oid == pd->oid) {
        return 2;
    }
    if (t->visible == 255 && (t->flags & FlagRoad) && ! t->oid) {
        return 1;
    }
    return 0;
}
int path_walk_road_or_grass_cb(PathData * pd, int x, int y) {
    Tile * t = map_get(x, y);
    if (t->oid == 0) {
        if (abs(pd->x2 - x) <= pd->close && abs(pd->y2 - y) <= pd->close) {
            return 2;
        }
    }
    else if (t->oid == pd->oid) {
        return 2;
    }
    if (t->visible == 255 && (t->flags & FlagDirt) == 0 && (t->flags & FlagWater) == 0 && t->oid == 0) {
        return 1;
    }
    return 0;
}
static PathNode * _global_nodes;
static int _global_nodes_count = 0;
int path_find(int x1, int y1, PathData * pd, int maxlen, int * dirs, int(* estimate_cb)(PathData *, int, int), int(* walk_cb)(PathData *, int, int)) {
    /* returns:
     * 0 -> no path found
     * 1 -> found path of length 1 (it was on a neighbor tile)
     * 2..n -> found path of length 2..n
     * (if the initial position is the target it will *not* be found
     * except perhaps as a path of length 2 to go away and return)
     */
    int n = game->map->w * game->map->h;
    // Node: on macos the main thread has about 8MB of stack but other
    // thread only have 512KB and there's no way to increase it
    if (maxlen > _global_nodes_count) {
        _global_nodes_count = maxlen;
        _global_nodes = land_realloc(_global_nodes, sizeof (PathNode) * maxlen);
    }
    PathNode * nodes = _global_nodes;
    char mark [n];
    memset(mark, 0, n);
    int used = 0;
    if (! _q) {
        _q = land_queue_new(_cmp_cb);
    }
    land_queue_clear(_q);
    PathNode * node0 = nodes + used++;
    node0->steps = 0;
    node0->x = x1;
    node0->y = y1;
    node0->d = - 1;
    node0->estimate = estimate_cb(pd, x1, y1);
    node0->parent = NULL;
    land_queue_add(_q, node0);
    while (! land_queue_is_empty(_q)) {
        PathNode * node = land_queue_pop(_q);
        for (int d = 0; d < 4; d += 1) {
            int x = node->x;
            int y = node->y;
            _move(& x, & y, d);
            if (mark [y * game->map->w + x] == 1) {
                continue;
            }
            int result = walk_cb(pd, x, y);
            if (result == 2) {
                // Found it!
                PathNode * trace = node;
                int i = trace->steps;
                dirs [i] = d;
                while (i > 0) {
                    i--;
                    dirs [i] = trace->d;
                    trace = trace->parent;
                }
                return node->steps + 1;
            }
            mark [y * game->map->w + x] = 1;
            if (result == 0) {
                continue;
            }
            PathNode * node1 = nodes + used;
            node1->steps = node->steps + 1;
            node1->x = x;
            node1->y = y;
            node1->d = d;
            node1->parent = node;
            node1->estimate = estimate_cb(pd, x, y);
            if (node1->steps + node1->estimate < maxlen) {
                land_queue_add(_q, node1);
                used++;
            }
            else {
                ;
            }
        }
    }
    //print("path exceeded %d (%d+%d)", maxlen, node1.steps, node1.estimate)
    return 0;
}
int path_find_object_near_road(int x1, int y1, int oid, int maxlen, int * steps) {
    PathData pd = {0, 0, oid, 0};
    return path_find(x1, y1, & pd, maxlen, steps, _estimate_unknown, path_walk_road_cb);
}
int path_find_position_on_road(int x1, int y1, int x2, int y2, int maxlen, int * steps) {
    PathData pd = {x2, y2, 0, 0};
    return path_find(x1, y1, & pd, maxlen, steps, _estimate_distance, path_walk_road_cb);
}
int path_find_close_position_on_road_or_grass(int x1, int y1, int x2, int y2, int maxlen, int close, int * steps) {
    PathData pd = {x2, y2, 0, close};
    return path_find(x1, y1, & pd, maxlen, steps, _estimate_distance, path_walk_road_or_grass_cb);
}
